!+
! Module: intmakehist_mod
!
! Authors:  M.T. Ceballos -  SSC (IFCA)
!           F.J. Carrera -  SSC (IFCA)
!           X. Barcons - SSC (IFCA)
!          
! Description: Creates a one-page PostScript File containig the histograms
!           for flatfielded EPIC Images in the default energy band
!              
! Comments: Flatfielding process previously carried out in Perl code INTHIST
! 
! $Id: intmakehist_mod.f90,v 1.2 1998/10/15 $
!-
module intmakehist_mod

  use dal
  use param  
  use errorHandling
  use sort_mod
  
  implicit none 
  
  private 

  public intmakehist

contains 

  !---------------------------------------------------------------------
  subroutine intmakehist
  !---------------------------------------------------------------------
    implicit none

    real(kind=single) :: binpixsize,inpixsize,x1,x2,y1,y2,expos,ypos
    real, dimension(:), allocatable :: histogram,NbinsInLastBin
    real(kind=single), dimension(:,:), pointer :: pr32_inImage
    real(kind=double), dimension(:,:), pointer :: p_inImage
    real(kind=double), dimension(:,:),allocatable :: binImage
    real(kind=double), dimension(:), allocatable :: ImageList,bin_value
    real(kind=double) :: mean,sigma,clip,binHistsize,mini,maxi,cutoff,&
         bin_low,bin_high

    integer :: pgbeg,nflats,pos1,pos2,pos3
    integer :: iband,i,j,nbinTot
    integer :: ib,jb,it,pixInbin,dimValid,nofpix,dim1_bin,dim2_bin
    integer :: all_stat,k,ih,type_inImage,ipgstat,nbins=10
    integer, dimension(:), pointer :: shape_inImage
    integer(kind=Int8),dimension(:,:), pointer :: pi8_inImage
    integer(kind=Int16),dimension(:,:), pointer :: pi16_inImage
    integer(kind=Int32),dimension(:,:), pointer :: pi32_inImage

    
    character(len=50)::infoLabel
    character(len=1000),dimension(:),allocatable :: inImage
    character(len=1000) :: outps,dev,filter
    character(len=25) :: band,low,upp,ibandchar

    logical :: alloc_ptr
    
    type(DataSetT) :: set_inImage
    type(ArrayT) :: arr_inImage

! ----------------------------------------------------------------
!   read input parameters from param file
! ----------------------------------------------------------------

    nflats=parameterCount("sets")

    allocate(inImage(nflats))
    do i=1,nflats
       inImage(i)=stringParameter("sets",i-1)
    enddo

    binpixsize=realParameter("binsize")
    
    outps=stringParameter("plotfile")    

    dev=trim(outps) // "/ps"
    ipgstat=pgbeg(0,trim(dev),1,1)
    if (ipgstat /= 1) call error('badDevice','Cannot open PGPLOT device')
    call pgsubp(3,2)

    print *,"Input parameters read"

! ----------------------------------------------------------------
!   input EPIC FITS flatfielded images for each band
! ----------------------------------------------------------------

    allocate(NbinsInLastBin(1:nflats))
    bands: do iband=1,nflats
       print *,"Iband=",iband
       set_inImage=dataSet(inImage(iband),Read)
       arr_inImage=array(set_inImage,"PRIMARY",Read)
       type_inImage=ArrayDataType(arr_inImage)
       shape_inImage=> dimensions(arr_inImage)

       if(hasAttribute(set_inImage,"SLCTEXPR"))then
          filter=stringAttribute(set_inImage,"SLCTEXPR")
       elseif(hasAttribute(arr_inImage,"SLCTEXPR"))then
          filter=stringAttribute(arr_inImage,"SLCTEXPR")
       else
          filter="none"
       endif
       if(filter=="none")then
          band="UNDEFINED"
       else
!   commented until SOC responds to SPR 2633 about stringAttribute          
          !extracting PI filter expression from SLCTEXPR
!          pos1=INDEX(filter,"PI>")
!          pos2=INDEX(filter,"&& PI<")
!          pos3=pos2+12
!          filter=filter(pos1:pos3)
!          pos2=INDEX(filter,"&& PI<")
!          pos2=pos2-1
!          low=trim(filter(4:pos2))
!          pos1=INDEX(filter,"PI<")
!          pos1=pos1+3
!          pos2=INDEX(filter,"00)")
!          pos2=pos2+1
!          upp=trim(filter(pos1:pos2))
!          band="Band ["//low//":"//upp//"]"
          write(ibandchar,*)iband
          band="Band number "//trim(ibandchar)
       endif
! ----------------------------------------------------------------
!      read input images  pixel size from CDELT1 WCS coordinate
! ----------------------------------------------------------------

       call message(AppMsg,Verbose,'Reading input images pixel size from CDELT1')
       
       if(hasAttribute(set_inImage,"CDELT1"))then
          inpixsize=real32Attribute(set_inImage,"CDELT1")
          inpixsize=ABS(inpixsize)
       elseif(hasAttribute(arr_inImage,"CDELT1"))then
          inpixsize=real32Attribute(arr_inImage,"CDELT1")
          inpixsize=ABS(inpixsize)
       else
          write(errstr,*)"No keyword CDELT1 in:",inImage(iband)
          call error('MissingFitsKeyw')
       endif

! ----------------------------------------------------------------
!      read input images exposure (the same for all)
! ----------------------------------------------------------------

       expos=0.0
       if(hasAttribute(set_inImage,"EXPOSURE"))then
          expos=real32Attribute(set_inImage,"EXPOSURE")
       elseif(hasAttribute(arr_inImage,"EXPOSURE"))then
          expos=real32Attribute(arr_inImage,"EXPOSURE")
       endif
! ----------------------------------------------------------------
!      check the data type of the input image and assign a pointer
! ----------------------------------------------------------------

       alloc_ptr=.TRUE.
       select case(type_inImage)
       case(INTEGER8)
          pi8_inImage => int8Array2Data(arr_inImage)
          allocate(p_inImage(0:shape_inImage(0)-1,0:shape_inImage(1)-1))
          p_inImage = real(pi8_inImage, DOUBLE)
       case(INTEGER16)
          pi16_inImage => int16Array2Data(arr_inImage)
          allocate(p_inImage(0:shape_inImage(0)-1,0:shape_inImage(1)-1))
          p_inImage = real(pi16_inImage, DOUBLE)
       case(INTEGER32)
          pi32_inImage => int32Array2Data(arr_inImage)
          allocate(p_inImage(0:shape_inImage(0)-1,0:shape_inImage(1)-1))
          p_inImage = real(pi32_inImage, DOUBLE)
       case(REAL32)
          pr32_inImage => real32Array2Data(arr_inImage)
          allocate(p_inImage(0:shape_inImage(0)-1,0:shape_inImage(1)-1))
          p_inImage = real(pr32_inImage, DOUBLE)
       case(REAL64)
          p_inImage => real64Array2Data(arr_inImage)
          alloc_ptr=.FALSE.
       case DEFAULT
          call fatal('InvalidDataType','Invalid data type in input Image')
       end select
       

! ----------------------------------------------------------------
!      Calculate the number of input pixels per binned pixel
! ----------------------------------------------------------------

       binpixsize=realParameter("binsize")

       if(binpixsize.lt.inpixsize)then
          call fatal('BadParamValue','Requested parameter binsize&
               & smaller than input image pixel size')
       endif
       pixInbin=FLOOR(binpixsize/inpixsize)  ! (dg/dg)
       
! ----------------------------------------------------------------
!      update the binned pixel size (since the value has been 
!      rounded to closest integer)
! ----------------------------------------------------------------

       binpixsize=inpixsize*pixInbin         ! (dg)

! ----------------------------------------------------------------
!      eliminate the possibility of a binned image with only 0 or 1 
!      pixel
! ----------------------------------------------------------------


       dim1_bin=FLOOR(REAL(shape_inImage(0))/REAL(pixInbin))
       dim2_bin=FLOOR(REAL(shape_inImage(1))/REAL(pixInbin))
       if(dim1_bin==0 .OR. dim2_bin==0)then
          call fatal('BadParamValue','Requested parameter binsize &
               & larger than input image size')
       elseif(dim1_bin==1 .AND. dim2_bin==1)then
          call fatal('BadParamValue','Requested parameter binsize &
               & results in a binned image with only one pixel')
       endif


! ----------------------------------------------------------------
!      Create temporal array to keep the rebinned image
! ----------------------------------------------------------------

       allocate(binImage(0:dim1_bin-1,0:dim2_bin-1), &
            stat=all_stat)
       if(all_stat /= 0)then
          call fatal('AllocationError','Failed Allocation for internal &
               & array binImage')
       endif

! ----------------------------------------------------------------
!      accumulate counts in each binned pixel of Binned Image:
!       a) "bad"  input pixels (ex. those with a corresponding exposure 
!          map value of 0) are given negative values
!       b) The addition is performed just over "good" pixels (>=0 counts)
!       c) nofpix = number of contributing pixels to each binned pixel
!       d) a normalisation is performed: if 3 out of 4 pixels contribute 
!          to the total counts, this total value is multiplied by 4/3
!       e) dimValid: number of "valid" binned pixels (>=0 counts)       
! ----------------------------------------------------------------
       call message(AppMsg,Verbose,'accumulating counts in each binned pixel of binned image')
       binImage=0.d0
       dimValid=0
       binImageX: do ib=0,dim1_bin-1
          binImageY: do jb=0,dim2_bin-1
             nofpix=0
             inImageX: do i=ib*pixInbin,(ib+1)*pixInbin-1
                inImageY: do j=jb*pixInbin,(jb+1)*pixInbin-1
                   if(p_inImage(i,j) >= 0) then
                      nofpix=nofpix+1
                      binImage(ib,jb)=binImage(ib,jb)+ &
                           p_inImage(i,j)
                   endif
                enddo inImageY
             enddo inImageX

             if (nofpix /= 0) then
                binImage(ib,jb)=binImage(ib,jb)*pixInbin**2./REAL(nofpix)
             else
                binImage(ib,jb)=-1.0d0
             endif
             if (binImage(ib,jb)>=0)then
                dimValid=dimValid+1
             endif
          
          enddo binImageY
       enddo binImageX


       call release(arr_inImage)
       call release(set_inImage)

       
! ----------------------------------------------------------------
!      allocate memory for List of "valid" Pixels (>=0)in Binned Image 
! ----------------------------------------------------------------

       allocate(ImageList(1:dimValid),stat=all_stat)
       if(all_stat /= 0)then
          call fatal('AllocationError','Failed Allocation for internal &
               & array ImageList')
       endif
       
! ----------------------------------------------------------------
!      Extract Binned-image "valid" pixels to a rank 1 array
!      and sort the array in increasing order
! ----------------------------------------------------------------

       k=0
       do ib=0,dim1_bin-1
          do jb=0,dim2_bin-1
             if(binImage(ib,jb) >= 0)then
                k=k+1
                ImageList(k)=binImage(ib,jb)
             endif
          enddo
       enddo

       call quickSort(ImageList)

! ----------------------------------------------------------------
!      Sigma-clipping process to decide the adequate bin size 
!      in the histogram bump:
!
!      1) Calculate mean, sigma, mean+3sigma in the "valid" pixels 
!         list
!      2) Eliminate from the list of "valid" pixels those with
!         more than mean+3sigma counts (brighter sources)
!      3) Go to 1) again
!      4) Go to 2) again
!      5) This will be repeated until no more pixels are eliminated
!         from "valid" pixels list.
! ----------------------------------------------------------------

       call message(AppMsg,Verbose,'Sigma-clipping process to decide bin size in histogram bump')

       CLIP_CALC:do
          mean=SUM(ImageList(1:dimValid))/dimValid
          sigma=sqrt(SUM((ImageList(1:dimValid)-mean)**2)/(dimValid-1))
          clip=mean+3*sigma
          
          if(ImageList(dimValid) > clip)then
             it=0
             LOWER: do 
                it=it+1
                if(it > dimValid)then
                   EXIT CLIP_CALC
                elseif (ImageList(it) > clip)then
                   EXIT LOWER
                endif
             enddo LOWER
             dimValid=it-1
          else
             EXIT CLIP_CALC
          endif
       enddo CLIP_CALC

! ----------------------------------------------------------------
!     Establish a cut-off: pixels with more than "cutoff" counts
!               will be accumulated in the last histogram bin
!               (bright sources)
! ----------------------------------------------------------------
      
       cutoff=mean+5*sigma
       write(errstr,*)'Cuttoff for last bin:',cutoff
       call message(AppMsg,Verbose)

! ----------------------------------------------------------------
!      Histogram Bin size calculation : nbins in the interval
!             between Minimum and Maximum value of pixels with
!             counts under the clipping value (mean + 3sigma)
! ----------------------------------------------------------------

       maxi=MAXVAL(ImageList(1:dimValid))
       mini=MINVAL(ImageList(1:dimValid))
       binHistsize=(maxi-mini)/REAL(nbins)

! ----------------------------------------------------------------
!      Build Histogram FROM the minimum Binned Image pixel value
!      ("mini") TO the selected cut-off ("cutoff) 
! ----------------------------------------------------------------

       call message(AppMsg,Verbose,'Building Histogram')

       if(binHistsize >0)then
          nbinTot=CEILING((cutoff-mini)/binHistsize)+1
          cutoff=mini+(nbinTot-1)*binHistsize

          allocate (histogram(1:nbinTot),stat=all_stat)
          if(all_stat /= 0)&
               call fatal('AllocationError','Failed Allocation for &
               & internal array histogram')

          allocate(bin_value(1:nbinTot),stat=all_stat)
          if(all_stat /= 0)&
               call fatal('AllocationError','Failed Allocation for &
               & internal array bin_value')

          do ih=1,nbinTot-1
             bin_low=mini+(ih-1)*binHistsize
             bin_high=bin_low + binHistsize
             bin_value(ih)=bin_low + binHistsize/2.0
          enddo
          bin_value(nbinTot)=bin_value(nbinTot-1)+binHistsize

          histogram=0
          do i=1,SIZE(ImageList)
             if(ImageList(i)>=cutoff)then
                ih=nbinTot
             else
                ih=INT((ImageList(i)-mini)/binHistsize)+1
             endif
             histogram(ih)=histogram(ih)+1
          enddo

! ----------------------------------------------------------------
!         PGPLOT creation of histogram in current band
! ----------------------------------------------------------------

          call message(AppMsg,Verbose,'PGPLOT PS file for histogram...')

          call pgpage
          call pgscf(2)
          call pgqvp(0,x1,x2,y1,y2)
          call pgsvp(x1,x2,y1*1.1,y2)
          call pgswin(real(mini),real(cutoff+binHistsize),0.,&
               MAXVAL(histogram)*1.1)
          call pgbox("BCNST",0.0,0,"BCNST",0.0,0)
          call pglab("countrate (ct s\u-1\d)",&
               "Number of bins",trim(band))
          call pgbin(nbinTot,real(bin_value),histogram,.TRUE.)
          call pgarro(real(bin_value(nbinTot)-binHistsize/2.),&
               histogram(nbinTot)/2.,&
               real(bin_value(nbinTot)+binHistsize/2.),&
               histogram(nbinTot)/2.)
          NbinsInLastBin(iband)=histogram(nbinTot)
          
          deallocate(histogram,stat=all_stat)
          if(all_stat /= 0)&
               call fatal('AllocationError','Failed DeAllocation for &
               & internal array histogram')
          deallocate(bin_value,stat=all_stat)
          if(all_stat /= 0)&
               call fatal('AllocationError','Failed DeAllocation for &
               & internal array bin_value')
       else
! ----------------------------------------------------------------
!         In case "binHistsize" <=0 : Plot a box warning the user 
!         and giving some information
! ----------------------------------------------------------------
          
          write(errstr,*)"All the pixels have the same intensity &
               &in flatfielded image in band:",band
          call warning('EmptyHistogram')
          call pgenv(0.,1.,0.,1.,1,-1)
          write(infoLabel,"("" Band: "",I2)")iband
          call pgtext(0.2,0.7,infoLabel)
          write(infoLabel,*)"All pixels have the same intensity"
          call pgtext(0.2,0.5,infoLabel)
          write(infoLabel,*)"No histogram plotted"
          call pgtext(0.2,0.3,infoLabel)
       endif
       deallocate(ImageList,stat=all_stat)
       if(all_stat /= 0)&
            call fatal('AllocationError','Failed DeAllocation for &
            & internal array ImageList')

       deallocate(binImage,stat=all_stat)
       if(all_stat /= 0)&
            call fatal('AllocationError','Failed DeAllocation for &
            & internal array binImage')

       if(alloc_ptr)then
          deallocate(p_inImage,stat=all_stat)
          if(all_stat /= 0)&
               call fatal('AllocationError','Failed DeAllocation for &
               & pointer p_inImage')
       endif
    enddo bands

! ----------------------------------------------------------------
!   After finishing plotting the histograms, add a new box with some 
!   comments: angular bin size, image exposure, etc.
! ----------------------------------------------------------------


    call message(AppMsg,Verbose,'Creatin last box with info...')

    call pgenv(0.,1.,0.,1.,1,-2)
    write(infoLabel,"("" Angular bin size: ("",F7.1,&
         &  "")\u2\d arcsec\u2\d"")") binpixsize*3600
    call pgtext(0.1,0.95,trim(infoLabel))
    write(infoLabel,"("" Exposure time: "",E12.5,"" sec"")")&
         expos
    call pgtext(0.1,0.85,trim(infoLabel))
    call pgtext(0.1,0.75," Number of bins in Last Bin (for each band):")
    do iband=1,nflats
       write(infoLabel,"('      Band',I2,':',F7.1)")iband,NbinsInLastBin(iband)
       ypos=0.75-0.05*iband
       call pgtext(0.1,ypos,trim(infoLabel))
    enddo
    call pgend
    
  end subroutine intmakehist
  
    
end module intmakehist_mod




